# Copyright (C) 2021, Camilo Hincapié Gutiérrez
# This file is part of CDSLIB.
#
# CDSLIB is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CDSLIB is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
#
# This package is authored by:
# Camilo Hincapié (https://www.linkedin.com/in/camilo-hincapie-gutierrez/) (main author)
# Ian Mejía (https://github.com/IanMejia)
# Emil Rueda (https://www.linkedin.com/in/emil-rueda-424012207/)
# Nicole Rivera (https://github.com/nicolerivera1)
# Carolina Rojas Duque (https://github.com/carolinarojasd)

"""Contagious diseases simulation using Agent-Based Models"""

__version__ = "0.0.4"

from abmodel import agent
from abmodel import models
from abmodel.analysis import Aggregator
from abmodel.population import Population
from abmodel.utils import Distribution
from abmodel.utils import ExecutionModes
from abmodel.utils import EvolutionModes


__all__ = [
    "agent",
    "models",
    "Aggregator",
    "Population",
    "Distribution",
    "ExecutionModes",
    "EvolutionModes"
    ]



# Copyright (C) 2021, Camilo Hincapié Gutiérrez
# This file is part of CDSLIB.
#
# CDSLIB is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CDSLIB is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
#
# This package is authored by:
# Camilo Hincapié (https://www.linkedin.com/in/camilo-hincapie-gutierrez/) (main author)
# Ian Mejía (https://github.com/IanMejia)
# Emil Rueda (https://www.linkedin.com/in/emil-rueda-424012207/)
# Nicole Rivera (https://github.com/nicolerivera1)
# Carolina Rojas Duque (https://github.com/carolinarojasd)

from enum import Enum
from typing import Optional, Any, Union
from datetime import datetime
from dataclasses import field
from datetime import timedelta

from pydantic import root_validator
from pydantic.dataclasses import dataclass
from numpy import random

from abmodel.models import SimpleGroups
from abmodel.utils.units import timedelta_to_days

# ============================================================================
# Mobility restrictions by tracing variables of interest
# (hereinafter abbreviated as MRT)
#
# Mobility restrictions (hereinafter abbreviated as MR or mr)
# ============================================================================


class InterestVariables(Enum):
    """
        This class enumerates the variables of interest
        for tracing and determining mobility restriction policies.
    """
    dead = "dead_by_disease"
    diagnosed = "diagnosed"
    ICU = "ICU_capacity"
    hospital = "hospital_capacity"


class MRTStopModes(Enum):
    """
        TODO: Add brief explanation
    """
    level_number = "level_number"
    length = "length"


class MRTimeUnits(Enum):
    """
        TODO: Add brief explanation
    """
    days = "days"
    weeks = "weeks"
    months = "months"


def time_interval_to_steps(
    mr_length: int,
    mr_length_units: MRTimeUnits,
    iteration_time: timedelta
) -> float:
    """
        TODO: Add brief explanation
    """
    iteration_time = timedelta_to_days(iteration_time)
    if mr_length_units == MRTimeUnits.days:
        mr_length_in_days = mr_length
    if mr_length_units == MRTimeUnits.weeks:
        mr_length_in_days = 7*mr_length
    if mr_length_units == MRTimeUnits.months:
        mr_length_in_days = 30*mr_length
    mr_length_in_steps = \
        timedelta_to_days(timedelta(days=mr_length_in_days)) / iteration_time
    return mr_length_in_steps


def random_time_interval_to_steps(
    mr_length: int,
    mr_length_units: MRTimeUnits,
    iteration_time: timedelta
) -> float:
    """
        TODO: Add brief explanation
    """
    iteration_time = timedelta_to_days(iteration_time)
    unrestricted_time = int(random.randint(1, mr_length + 1, 1)[0])

    if mr_length_units == MRTimeUnits.days:
        unrestricted_time_in_days = unrestricted_time
    if mr_length_units == MRTimeUnits.weeks:
        unrestricted_time_in_days = 7*unrestricted_time
    if mr_length_units == MRTimeUnits.months:
        unrestricted_time_in_days = 30*unrestricted_time
    unrestricted_time_steps = \
            timedelta_to_days(
                timedelta(days=unrestricted_time_in_days)
                ) / iteration_time
    return unrestricted_time_steps


@dataclass
class MRTracingPolicies:
    """
        TODO: Add brief explanation

        Attributes
        ----------
        TODO

        Examples
        --------
        TODO: include some examples
    """
    variable: InterestVariables
    mr_start_level: int
    mr_stop_mode: MRTStopModes
    mr_groups: SimpleGroups
    target_groups: list[str]
    mr_stop_level: Optional[int] = None
    mr_length: Optional[int] = None
    mr_length_units: Optional[MRTimeUnits] = None
    mr_length_in_steps: float = field(default=None, init=False)

    @root_validator
    def validate(
        cls,
        v: dict[str, Any]
    ) -> dict[str, Any]:
        """
            TODO
        """
        mode = v.get("mr_stop_mode")

        if mode == MRTStopModes.level_number:
            if not v.get("mr_stop_level"):
                raise ValueError(
                    "A valid value for `mr_stop_level` was expected"
                    )
            if v.get("mr_length") and v.get("mr_length_units"):
                raise ValueError(
                    f"""
                        `mr_stop_mode` was set to
                        `{MRTStopModes.level_number}`.
                        So, `mr_length` and `mr_length_units`
                        should not be provided.
                    """
                    )
        if mode == MRTStopModes.length:
            if not v.get("mr_length"):
                raise ValueError(
                    "A valid value for `mr_length` was expected"
                    )
            if not v.get("mr_length_units"):
                raise ValueError(
                    "A valid value for `mr_length_units` was expected"
                    )
            if v.get("mr_stop_level"):
                raise ValueError(
                    f"""
                        `mr_stop_mode` was set to
                        `{MRTStopModes.length}`.
                        So, `mr_stop_level` should not be provided.
                    """
                    )
        # Check if all elements of `target_groups` are in `mr_groups`
        mr_groups = v.get("mr_groups").names
        target_groups = v.get("target_groups")
        if not all(elem in mr_groups for elem in target_groups):
            raise ValueError(
                f"""
                    All `target_groups` must be in
                    `mr_groups = {mr_groups}`.
                """
                )
        return v

    def set_mr_length_in_steps(self, iteration_time: timedelta):
        self.mr_length_in_steps = time_interval_to_steps(
            self.mr_length,
            self.mr_length_units,
            iteration_time
        )

# ============================================================================
# Cyclic mobility restrictions (hereinafter abbreviated as CyclicMR)
# ============================================================================


class CyclicMRModes(Enum):
    """
        TODO: Add brief explanation
    """
    random = "random"
    fixed = "fixed"


@dataclass
class GlobalCyclicMR:
    """
        TODO: Add brief explanation

        Attributes
        ----------
        TODO

        Examples
        --------
        TODO: include some examples
    """
    enabled: bool
    grace_time: datetime
    global_mr_length: int
    global_mr_length_units: MRTimeUnits
    unrestricted_time_mode: CyclicMRModes
    unrestricted_time_units: MRTimeUnits
    unrestricted_time: Optional[int] = None
    global_mr_length_steps: float = field(default=False, init=False)
    unrestricted_time_steps: Union[float, None] = None

    @root_validator
    def validate(
        cls,
        v: dict[str, Any]
    ) -> dict[str, Any]:
        """
            TODO
        """
        mode = v.get("unrestricted_time_mode")

        if mode == CyclicMRModes.random and v.get("unrestricted_time"):
            raise ValueError(
                f"""
                    `unrestricted_time_mode` was set to
                    `{CyclicMRModes.random}`.
                    So, `unrestricted_time` should not be provided.
                """
                )
        if mode == CyclicMRModes.fixed and not v.get("unrestricted_time"):
            raise ValueError(
                f"""
                    `unrestricted_time_mode` was set to
                    `{CyclicMRModes.fixed}`.
                    So, `unrestricted_time` should be provided.
                """
                )
        return v

    def set_global_mr_length(self, iteration_time: timedelta):
        self.global_mr_length_steps = time_interval_to_steps(
            self.global_mr_length,
            self.global_mr_length_units,
            iteration_time
        )

    def set_unrestricted_time(self, itertation_time: timedelta):
        if self.unrestricted_time_mode == CyclicMRModes.fixed:
            self.unrestricted_time_steps = time_interval_to_steps(
                self.unrestricted_time,
                self.unrestricted_time_units,
                itertation_time
            )
        if self.unrestricted_time_mode == CyclicMRModes.random:
            self.unrestricted_time_steps = \
                random_time_interval_to_steps(
                    self.global_mr_length,
                    self.unrestricted_time_units,
                    itertation_time
                )

    def set_none_unrestricted_time(self):
        self.unrestricted_time_steps = None


@dataclass
class CyclicMRPolicies:

    """
        TODO: Add brief explanation

        Attributes
        ----------
        TODO

        Examples
        --------
        TODO: include some examples
    """
    mr_groups: SimpleGroups
    target_group: str
    delay: int
    delay_units: MRTimeUnits
    mr_length: int
    mr_length_units: MRTimeUnits
    time_without_restrictions: int
    time_without_restrictions_units: MRTimeUnits
    delay_in_steps: float = field(default=False, init=False)
    mr_length_in_steps: float = field(default=False, init=False)
    time_without_restrictions_steps: float = \
        field(default=False, init=False)

    def set_delay(self, iteration_time: timedelta):
        self.delay_in_steps = time_interval_to_steps(
        self.delay,
        self.delay_units,
        iteration_time
    )

    def set_mr_length(self, iteration_time: timedelta):
        self.mr_length_in_steps = time_interval_to_steps(
            self.mr_length,
            self.mr_length_units,
            iteration_time
        )

    def set_time_without_restrictions(
        self,
        iteration_time: timedelta
    ):
        self.time_without_restrictions_steps = time_interval_to_steps(
            self.time_without_restrictions,
            self.time_without_restrictions_units,
            iteration_time
        )



# Copyright (C) 2021, Camilo Hincapié Gutiérrez
# This file is part of CDSLIB.
#
# CDSLIB is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CDSLIB is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
#
# This package is authored by:
# Camilo Hincapié (https://www.linkedin.com/in/camilo-hincapie-gutierrez/) (main author)
# Ian Mejía (https://github.com/IanMejia)
# Emil Rueda (https://www.linkedin.com/in/emil-rueda-424012207/)
# Nicole Rivera (https://github.com/nicolerivera1)
# Carolina Rojas Duque (https://github.com/carolinarojasd)

from typing import Callable

from pandas.core.frame import DataFrame
from pandas import concat

from abmodel.utils import ExecutionModes


class Aggregator:
    """
        TODO: Add brief explanation

        Attributes
        ----------
        TODO

        Methods
        -------
        TODO
    """
    def __init__(
        self,
        col_name: str,
        col_values: list,
        read_method: Callable,
        prefix: str = "sim_",
        execmode: ExecutionModes = ExecutionModes.iterative.value
    ) -> None:
        """
            Constructor of Population class.

            TODO: Add brief explanation

            Parameters
            ----------
            TODO
        """
        # Store configuration
        self.col_name = col_name
        self.col_values = col_values
        self.read_method = read_method
        self.prefix = prefix
        self.execmode = execmode

    def __step_count_func(
        self,
        df: DataFrame,
        simulation_number: int,
    ) -> DataFrame:
        """
            TODO
        """
        grouped_df = df[[self.col_name, "agent"]] \
            .groupby(self.col_name) \
            .count().rename(
                columns={"agent": "_".join([self.prefix, simulation_number])}
            )

        grouped_df = grouped_df.reindex(self.col_values, fill_value=0)

        return grouped_df

    def aggregation_by_step(
        self,
        total_steps: int,
        filepaths_dict: dict
    ) -> DataFrame:
        """
            TODO
        """
        # For storing all aggregated df (one by step)
        all_steps_df = []

        for step in range(total_steps):

            filepaths = filepaths_dict[step]

            df_list = [self.read_method(filepath) for filepath in filepaths]

            # datetime = df_list[0]["datetime"][0]

            # df_list = [
            #     self.__step_count_func(df=df, simulation_number=i)
            #     for i, df in enumerate(df_list)
            #     ]

            df_list = [
                self.__count_func(df=df, simulation_number=i)
                for i, df in enumerate(df_list)
                ]

            df = concat(df_list, axis=1)

            # df = df.assign(step=step)
            # df = df.assign(datetime=datetime)

            # df.reset_index(inplace=True)

            all_steps_df.append(df)

        df = concat(all_steps_df)

        df = df.set_index(["step", "datetime", self.col_name])

        df.insert(0, "mean", df.mean(axis=1))
        df.insert(1, "std", df.std(axis=1))
        df.insert(2, "min", df.min(axis=1))
        df.insert(3, "max", df.max(axis=1))

        df.reset_index(inplace=True)

        return df

    def __count_func(
        self,
        df: DataFrame,
        simulation_number: int,
    ) -> DataFrame:
        """
            TODO
        """
        index_cols = ["step", "datetime"]
        cols = index_cols + [self.col_name]
        target = "agent"
        label = "_".join([self.prefix, simulation_number])

        partial_df = df[cols + [target]] \
            .groupby(cols) \
            .count().rename(
                columns={target: label}
            ).reset_index()

        agg_df = partial_df.groupby(index_cols).apply(
            lambda grouped_df: grouped_df[[self.col_name, label]]
            .set_index(self.col_name)
            .reindex(self.col_values, fill_value=0)
            .reset_index()
            ).reset_index()[cols + [label]]

        return agg_df

    def aggregation_of_cumulatives(
        self,
        filepaths: list
    ) -> DataFrame:
        """
            TODO
        """
        df_list = [self.read_method(filepath) for filepath in filepaths]

        df_list = [
            self.__count_func(df=df, simulation_number=i)
            for i, df in enumerate(df_list)
            ]

        df = concat(df_list)

        df = df.set_index(["step", "datetime", self.col_name])

        df.insert(0, "mean", df.mean(axis=1))
        df.insert(1, "std", df.std(axis=1))
        df.insert(2, "min", df.min(axis=1))
        df.insert(3, "max", df.max(axis=1))

        df.reset_index(inplace=True)

        return df



# Copyright (C) 2021, Camilo Hincapié Gutiérrez
# This file is part of CDSLIB.
#
# CDSLIB is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CDSLIB is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
#
# This package is authored by:
# Camilo Hincapié (https://www.linkedin.com/in/camilo-hincapie-gutierrez/) (main author)
# Ian Mejía (https://github.com/IanMejia)
# Emil Rueda (https://www.linkedin.com/in/emil-rueda-424012207/)
# Nicole Rivera (https://github.com/nicolerivera1)
# Carolina Rojas Duque (https://github.com/carolinarojasd)

from numpy import full
from numpy.random import choice
from pandas.core.frame import DataFrame

from abmodel.models import SimpleGroups


class InitialArrangement:
    """
        TODO: Add brief explanation

        Methods
        -------
        TODO
    """
    @classmethod
    def setup(
        cls,
        df: DataFrame,
        core_var: str,
        nested_vars: list,
        settings: dict
    ) -> DataFrame:
        """
            I take a json-like input, a df contaning `agent` data and
            modify such df with the necessary data inside the json-like input.
            TODO: Improve explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Examples
            --------
            TODO: include some examples
        """
        if not nested_vars:
            # Get values and probabilities
            values = list(settings.keys())
            probabilities = list(settings.values())

            # All values are assigned randomly with the given probabilities
            df[core_var] = choice(a=values, p=probabilities, size=df.shape[0])
        else:
            # Create the new column with assigned value None
            df[core_var] = full(df.shape[0], None)

            # aux_df is going to have all nested_vars values
            aux_df = df[nested_vars].drop_duplicates()

            for _, row in aux_df.iterrows():
                # Retrieve (keys, items) from the row
                row_keys = list(row.keys())
                row_items = row.to_list()

                # Use row_items to get inner_settings dict
                inner_settings = settings
                for item in row_items:
                    inner_settings = inner_settings[item]

                # Get values and probabilities
                values = list(inner_settings.keys())
                probabilities = list(inner_settings.values())

                # Generate condition_list for filtering
                condition_list = [
                    f'(df["{key}"] == "{item}")'
                    for (key, item) in zip(row_keys, row_items)
                    ]

                # Get filtering condition
                _condition = ' & '.join(condition_list)

                # Assign values
                # All values are assigned randomly with the given probabilities
                df.loc[eval(_condition), core_var] = \
                    choice(
                        a=values,
                        p=probabilities,
                        size=df[eval(_condition)].shape[0]
                        )

        return df

    @classmethod
    def fulfill_setup(
        cls,
        df: DataFrame,
        missing_cols: list[str],
        req_cols_dict: dict
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Examples
            --------
            TODO: include some examples
        """
        for group_col in missing_cols:
            if req_cols_dict[group_col] is not None:
                if isinstance(req_cols_dict[group_col], SimpleGroups):
                    group_values = req_cols_dict[group_col].names
                else:
                    group_values = list(req_cols_dict[group_col].items.keys())

                # Fill group_values randomly
                df[group_col] = choice(
                    a=group_values,
                    size=df.shape[0]
                    )
            else:
                # req_cols_dict[group_col] is None

                # Fill group_values with None
                df[group_col] = None

        return df




# Copyright (C) 2021, Camilo Hincapié Gutiérrez
# This file is part of CDSLIB.
#
# CDSLIB is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CDSLIB is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
#
# This package is authored by:
# Camilo Hincapié (https://www.linkedin.com/in/camilo-hincapie-gutierrez/) (main author)
# Ian Mejía (https://github.com/IanMejia)
# Emil Rueda (https://www.linkedin.com/in/emil-rueda-424012207/)
# Nicole Rivera (https://github.com/nicolerivera1)
# Carolina Rojas Duque (https://github.com/carolinarojasd)

from typing import Union, Any
from dataclasses import dataclass, field
from copy import deepcopy

from munch import Munch
from pydantic import BaseModel

from abmodel.utils import Distribution
from abmodel.utils import init_distribution


# ============================================================================
# Groups with only names in their attributes
# ============================================================================

class SimpleGroups(BaseModel):
    """
        This dataclasss is best suited for supporting
        groups for which only their names are needed
        in the simulation.

        In this version of the library, it is used for:
        `age_groups`, `vulnerability_groups` and
        `mr_groups` (i.e. mobility restrictions groups).

        Attributes
        ----------
        names : list

        Examples
        --------
        TODO: include some examples
    """
    names: list


# ============================================================================
# Groups with distributions in their attributes
# ============================================================================

@dataclass
class DistributionGroup:
    """
        This dataclass uses `init_distribution` function
        in order to encapsulate a Distribution
        as the attribute `dist` belonging to this class.

        Attributes
        ----------
        name : str
            Group name.

        dist_info : dict or list of dict
            Dictionary with the required information
            in order to initialize the distribution.
            It must contain `dist_title` key, which refers to
            the distribution title.

        dist : dict of Distribution
            Distribution used in this group.

        Raises
        ------
            TODO

        See Also
        --------
        abmodel.utils.helpers.distributions.init_distribution : Function to
        initialize a distribution

        abmodel.utils.distributions.Distribution : Distribution class

        Examples
        --------
        TODO: include some examples
    """
    name: str
    dist_info: Union[dict[str, Any], list[dict[str, Any]]]
    dist: dict[Distribution] = field(init=False)

    def __post_init__(self):
        """
            This method initializes `dist` distribution
            dictionary from `dist_info` info.
        """
        def prepare_dict_field(dist_dict) -> None:
            """
                This method adds a single distribution
                to `dist` distribution dictionary

                Parameters
                ----------
                dist_dict : dict
                    It is a single distribution dictionary
            """
            dist_title = dist_dict.pop("dist_title")
            self.dist[dist_title] = init_distribution(dist_dict)

        # Initialize void dict
        self.dist = {}

        if isinstance(self.dist_info, dict):
            prepare_dict_field(self.dist_info)
        elif isinstance(self.dist_info, list):
            # self.dist_info is a list of dictionaries
            for dist_dict in self.dist_info:
                prepare_dict_field(dist_dict)
        else:
            # TODO
            raise ValueError("TODO")


@dataclass
class SimpleDistGroups:
    """
        This dataclass wraps a set of DistributionGroup
        in a dictionary in order to make easier
        their retrieval.

        Attributes
        ----------
        dist_title : str or list of str


        group_info : list of dict
            The list of different single group
            information required to instantiate
            a DistributionGroup.

        items : dict
            The dictionary created from the list `group_info`.
            Each key of this dictionary corresponds to a single
            group name.

        See Also
        --------
        DistributionGroup : Single Distribution group

        ComplexDistGroups : Set of groups compound
            by distributions and other additional fields.

        Examples
        --------
        TODO: include some examples
    """
    dist_title: Union[str, list[str]]
    group_info: list[dict]
    items: dict = field(init=False)

    def __post_init__(self):
        """
            This method performs `items` dictionary
            assignment from `group_info` list.
        """
        self.items = {
            single_group["name"]: DistributionGroup(**single_group)
            for single_group in self.group_info
        }

    def single_dist_title_validation(self, expected_dist_title: str):
        """
            Validates `dist_title` to be equal to
            `expected_dist_title`.

            Parameters
            ----------
            expected_dist_title : str
                Canonical value of the expected distribution
                title.

            Raises
            ------
            ValueError
                If `dist_title` is not equal to `expected_dist_title`.
                If `dist_title` is not the same for all the groups.
                If `dist_title` for all the groups is not equal to
                `expected_dist_title`.
        """
        # Assess if dist_title == expected_dist_title
        condition = self.dist_title == expected_dist_title
        assert condition, ValueError(
            "'dist_title' is not equal to "
            f"'{expected_dist_title}'"
            )

        # Retrieve dist_title list from all group's single dist_title
        dist_title_list = [
            group['dist_info']['dist_title']
            for group in self.group_info
            ]

        # Assess if dist_title is the same for all the groups
        condition = len(set(dist_title_list)) == 1
        assert condition, ValueError(
            f"The list of dist_title in the group is: {dist_title_list}"
            "\n"
            "All of them must be the same and equal to "
            f"'{expected_dist_title}'"
        )

        # Assess if dist_title for all the groups is equal to
        # expected_dist_title
        condition = dist_title_list[0] == expected_dist_title
        assert condition, ValueError(
            f"The dist_title for each group is: {dist_title_list[0]}"
            "\n"
            f"It must be equal to '{expected_dist_title}'"
        )

    def multiple_dist_title_validation(self, expected_dist_title):
        """
            TODO:
                - Update expected_dist_title type
                - Update this docstring

            Validates `dist_title` list to be equal to
            `expected_dist_title` list.

            Parameters
            ----------
            expected_dist_title : list of str
                Canonical list of values of the expected distribution
                titles.

            Raises
            ------
            ValueError
                If `dist_title` and `expected_dist_title` does not have the
                same length.

                If `dist_title` is the same for all the groups
                If `dist_title` for all the groups is equal to
                `expected_dist_title`
        """
        # Assess if dist_title and expected_dist_title have the same length
        condition = len(self.dist_title) == len(expected_dist_title)
        assert condition, ValueError(
            "'dist_title' does not have the same length of "
            f"'{expected_dist_title}'"
            )

        # Assess if dist_title and expected_dist_title have the same elements
        condition = set(self.dist_title) == set(expected_dist_title)
        assert condition, ValueError(
            "'dist_title' does not have the same elements of "
            f"'{expected_dist_title}'"
            )

        # Retrieve dist_title list from all group's dist_titles
        dist_title_list = [
            [dist['dist_title'] for dist in group['dist_info']]
            for group in self.group_info
            ]

        print(dist_title_list)
        # TODO: to finish validations
        # # Assess if dist_title is the same for all the groups
        # condition = len(set(dist_title_list)) == 1
        # assert condition, ValueError(
        #     f"The list of dist_title in the group is: {dist_title_list}"
        #     "\n"
        #     "All of them must be the same and equal to "
        #     f"'{expected_dist_title}'"
        # )

        # # Assess if dist_title for all the groups is equal to
        # # expected_dist_title
        # condition = dist_title_list[0] == expected_dist_title
        # assert condition, ValueError(
        #     f"The dist_title for each group is: {dist_title_list[0]}"
        #     "\n"
        #     f"It must be equal to '{expected_dist_title}'"
        # )


@dataclass
class ComplexDistGroups(SimpleDistGroups):
    """
        This dataclass wraps a set of DistributionGroup
        in a dictionary in order to make easier
        their retrieval, nevertheless it differs from `SimpleDistGroups`
        in the fact that in addition to the distribution,
        it is also intended to wrap information
        from other fields specified by `labels` list.

        Attributes
        ----------
        group_info : list of dict
            The list of different single group
            information required to instantiate
            a DistributionGroup.

        labels : list
            Labels of the other fields apart from `dist`
            distribution.

        items : dict
            The dictionary created from the list `group_info`
            and the list of `labels`. Each key of this
            dictionary corresponds to a single
            group name.

        See Also
        --------
        DistributionGroup : TODO complete explanation

        SimpleDistGroups : Set of groups compound only
            by distributions.

        Examples
        --------
        TODO: include some examples
    """
    labels: list = field(init=False)

    def __post_init__(self):
        """
            This method performs `items` dictionary
            assignment from `group_info` list.
        """
        # Copy self.group_info in order to avoid
        # that it gets modified
        group_info = deepcopy(self.group_info)

        self.items = {
            single_group["name"]: Munch({
                label.replace(" ", "_"): single_group.pop(label)
                for label in self.labels
                })
            for single_group in group_info
        }

        for single_group in group_info:
            self.items[
                single_group["name"]
                ].dist = deepcopy(
                    DistributionGroup(**single_group).dist
                )



# Copyright (C) 2021, Camilo Hincapié Gutiérrez
# This file is part of CDSLIB.
#
# CDSLIB is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CDSLIB is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
#
# This package is authored by:
# Camilo Hincapié (https://www.linkedin.com/in/camilo-hincapie-gutierrez/) (main author)
# Ian Mejía (https://github.com/IanMejia)
# Emil Rueda (https://www.linkedin.com/in/emil-rueda-424012207/)
# Nicole Rivera (https://github.com/nicolerivera1)
# Carolina Rojas Duque (https://github.com/carolinarojasd)

from .base import SimpleGroups
from .base import DistributionGroup
from .base import SimpleDistGroups
from .base import ComplexDistGroups
from .disease import DistTitles
from .disease import SusceptibilityGroups
from .disease import ImmunizationGroups
from .disease import MobilityGroups
from .disease import IsolationAdherenceGroups
from .disease import MRAdherenceGroups
from .disease import DiseaseStates
from .disease import Transitions
from .disease import NaturalHistory
from .health_system import HealthSystem
from .mobility_restrictions import InterestVariables
from .mobility_restrictions import MRTStopModes
from .mobility_restrictions import MRTimeUnits
from .mobility_restrictions import MRTracingPolicies
from .mobility_restrictions import CyclicMRModes
from .mobility_restrictions import GlobalCyclicMR
from .mobility_restrictions import CyclicMRPolicies
from .population import BoxSize
from .population import Configutarion


__all__ = [
    "SimpleGroups",
    "DistributionGroup",
    "SimpleDistGroups",
    "ComplexDistGroups",
    "DistTitles",
    "SusceptibilityGroups",
    "ImmunizationGroups",
    "MobilityGroups",
    "IsolationAdherenceGroups",
    "MRAdherenceGroups",
    "DiseaseStates",
    "Transitions",
    "NaturalHistory",
    "HealthSystem",
    "InterestVariables",
    "MRTStopModes",
    "MRTimeUnits",
    "MRTracingPolicies",
    "CyclicMRModes",
    "GlobalCyclicMR",
    "CyclicMRPolicies",
    "BoxSize",
    "Configutarion",
    ]



# Copyright (C) 2021, Camilo Hincapié Gutiérrez
# This file is part of CDSLIB.
#
# CDSLIB is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CDSLIB is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
#
# This package is authored by:
# Camilo Hincapié (https://www.linkedin.com/in/camilo-hincapie-gutierrez/) (main author)
# Ian Mejía (https://github.com/IanMejia)
# Emil Rueda (https://www.linkedin.com/in/emil-rueda-424012207/)
# Nicole Rivera (https://github.com/nicolerivera1)
# Carolina Rojas Duque (https://github.com/carolinarojasd)

from typing import Optional, Union

from math import fmod
from numpy import ndarray, arctan2, cos, sin, pi, sqrt, inf, frompyfunc, array
from pandas.core.frame import DataFrame, Series

from abmodel.models.population import BoxSize
from abmodel.models.disease import MobilityGroups, DistTitles
from abmodel.utils.distributions import Distribution
from abmodel.utils.utilities import check_field_errors
from abmodel.utils.utilities import check_field_existance, exception_burner


def move_individual_agent(
    row: Series,
    box_size: BoxSize,
    dt: float
) -> Series:
    """
    TODO: Add brief explanation
    A row represents an agent.

    Parameters
    ----------
    TODO

    Returns
    -------
    TODO

    Notes
    -----
    TODO: include mathematical description and explanatory image

    Examples
    --------
    TODO: include some examples
    """
    # Update current position of the agent with its velocities
    row.x += row.vx * dt
    row.y += row.vy * dt

    # Verify if coordinates are out of the box
    # then return to the box limit
    if row.x < box_size.left:
        row.vx = -row.vx
        row.x = box_size.left
    if row.x > box_size.right:
        row.vx = -row.vx
        row.x = box_size.right

    if row.y < box_size.bottom:
        row.vy = -row.vy
        row.y = box_size.bottom
    if row.y > box_size.top:
        row.vy = -row.vy
        row.y = box_size.top

    return row


class AgentMovement:
    """
        TODO: Add brief explanation

        Methods
        -------
        TODO
    """
    @classmethod
    def init_required_fields(
        cls,
        df: DataFrame,
        box_size: BoxSize,
        mobility_groups: MobilityGroups
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            move_individual_agent : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        # Initialize positions
        df["x"] = Distribution(
            dist_type="numpy",
            dist_name="uniform",
            low=box_size.left,
            high=box_size.right
            ).sample(size=df.shape[0])

        df["y"] = Distribution(
            dist_type="numpy",
            dist_name="uniform",
            low=box_size.bottom,
            high=box_size.top
            ).sample(size=df.shape[0])

        # Initialize velocities
        df = df.assign(vx=inf)
        df = df.assign(vy=inf)

        for mobility_group in mobility_groups.items.keys():
            df = cls.initialize_velocities(
                df=df,
                distribution=mobility_groups.items[
                    mobility_group].dist[DistTitles.mobility.value],
                angle_distribution=Distribution(
                    dist_type="numpy",
                    dist_name="uniform",
                    low=0.0,
                    high=2*pi
                    ),
                group_field="mobility_group",
                group_label=mobility_group,
                preserve_dtypes_dict={"step": int, "agent": int}
                )

        return df

    @classmethod
    def move_agents(
        cls,
        df: DataFrame,
        box_size: BoxSize,
        dt: float  # In scale of the mobility_profile
    ) -> DataFrame:
        """
            Function to apply as transformation in a pandas Dataframe to update
            coordinates from the agent with its velocities.

            Parameters
            ----------
            df : DataFrame
                Dataframe to apply transformation.
                Must have `x`, `y`, `vx` and `vy` columns.

            box_size : BoxSize
                Parameter with the region coordinates.

            dt : float
                Local time step, representing how often to take a measure.

            Returns
            -------
            df: DataFrame
                Dataframe with the transformations in columns `x` and `y`

            Raises
            ------
            ValueError
                If the dataframe `df` doesn't have `x`, `y`, `vx`
                and `vy` columns.

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            move_individual_agent : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        check_field_errors(df[["x", "y", "vx", "vy"]])
        try:
            df = df.apply(
                lambda row: move_individual_agent(row, box_size, dt),
                axis=1
                )

        except Exception as error:
            exception_burner([
                error,
                check_field_existance(df, ["x", "y", "vx", "vy"])
                ])
        else:
            return df

    @classmethod
    def stop_agents(
        cls,
        df: DataFrame,
        indexes: Union[list, ndarray]
    ) -> DataFrame:
        """
            Set the velocity of a given set of agents to zero.

            Parameters
            ----------
            df : DataFrame
                Dataframe to apply transformation.
                Must have `vx` and `vy` columns.

            indexes : list
                List containing the index of the agents that need to be
                stopped

            Returns
            -------
            df : DataFrame
                Dataframe with the transformations in columns `vx` and `vy`

            Raises
            ------
            ValueError
                If the dataframe `df` doesn't have `vx`
                and `vy` columns.

            Examples
            --------
            TODO: include some examples
        """
        check_field_existance(df, ["vx", "vy"])
        df.loc[indexes, "vx"] = 0
        df.loc[indexes, "vy"] = 0

        return df

    @classmethod
    def standardize_angle(cls, angle: float) -> float:
        """
            Standardize angles to be in the interval [-pi, pi]

            Parameters
            ----------
            angle : float
                The angle to be standardized

            Returns
            -------
            standardized_angle : float
                The standardized angle

            Notes
            -----
            TODO: include mathematical description and explanatory image

            Examples
            --------
            TODO: include some examples
        """
        return fmod(angle + 2*pi, 2*pi)

    @classmethod
    def angle(cls, x: float, y: float) -> float:
        """
            Returns the standardized angle formed by the components
            `x` and `y`.

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            standardize_angle : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        # Standardize angles on the interval [0, 2*pi]
        return cls.standardize_angle(arctan2(y, x))

    @classmethod
    def vector_angles(cls, df: DataFrame, components: list) -> Series:
        """
            Calculates vector angles from their euclidean components

            Parameters
            ----------
            df : DataFrame
                Dataframe with vector components to calculate the
                corresponding angles

            components: list
                Vector components names.
                If the vectors corresponds to positions, then
                `components = ['x', 'y']`.
                If the vectors corresponds to velocities, then
                `components = ['vx', 'vy']`.

            Returns
            -------
            angles : Series
                Serie with the computed angles.

            Raises
            ------
            ValueError
                If the dataframe `df` doesn't have the columns
                specified by `components`.

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            angle : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            angles = df.apply(
                lambda row: cls.angle(row[components[0]], row[components[1]]),
                axis=1
                )
        except Exception as error:
            exception_burner([
                error,
                check_field_existance(df, components)
                ])
        else:
            return angles

    @classmethod
    def set_velocities(
        cls,
        df: DataFrame,
        distribution: Distribution,
        angle_variance: Optional[float] = None,
        angle_distribution: Optional[Distribution] = None
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            Exception
                TODO complete explanation

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.utils.distributions.Distribution : TODO complete
            explanation

            standardize_angle : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            check_field_existance(df, ["vx", "vy"])
            n_agents = len(df.index)
            new_velocities_norm = distribution.sample(size=n_agents)

            if angle_distribution is None:
                # Use former angles as baseline to create new ones but modified
                # by a normal distribution of scale equal to angle_variance
                angles = cls.vector_angles(df, ["vx", "vy"])

                delta_angles = Distribution(
                    dist_type="numpy",
                    dist_name="normal",
                    loc=0.0,
                    scale=angle_variance
                    ).sample(size=n_agents)

                angles = angles + delta_angles

                # Standardize angles on the interval [0, 2*pi]
                angles = angles.apply(cls.standardize_angle)
            else:
                # Use angle_distribution to create new angles
                # This option is used for initializing velocities
                angles = angle_distribution.sample(size=n_agents)

                # Standardize angles on the interval [0, 2*pi]
                standardize_angle_array = frompyfunc(
                    cls.standardize_angle,
                    nin=1,
                    nout=1
                    )
                angles = array(standardize_angle_array(angles)).astype(float)

            df.loc[df.index, "vx"] = new_velocities_norm * cos(angles)

            df.loc[df.index, "vy"] = new_velocities_norm * sin(angles)

        except Exception as error:
            exception_burner([
                error,
                check_field_existance(df, ["vx", "vy"])
                ])
        else:
            return df

    @classmethod
    def initialize_velocities(
        cls,
        df: DataFrame,
        distribution: Distribution,
        angle_distribution: Distribution,
        indexes: Union[list, ndarray, None] = None,
        group_field: Optional[str] = None,
        group_label: Optional[str] = None,
        preserve_dtypes_dict: Optional[dict] = None
    ) -> DataFrame:
        """
            Initialize the velocity of a given set of agents from a given
            mobility profile (i.e. a velocity distribution) and ...
            TODO

            Parameters
            ----------
            df : DataFrame
                Dataframe to apply transformation.
                Must have `vx` and `vy` columns.

            distribution : Distribution
                Mobility profile. This is the velocity distribution
                to use for updating the population velocities each
                time step.

            angle_distribution : Distribution
                TODO

            indexes : TODO

            group_field : str, optional
                The field over which to filter the set of agents.
                If not provided, then the set of agents used is
                going to be the whole set of agents.

            group_label : str, optional
                The value of the `group_filed` used to filter
                the set of agents. If `group_field` is not provided,
                then this parameter is ignored.

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: insert mathematical description and explanatory image

            See Also
            --------
            set_velocities : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        if indexes is not None and group_field is not None:
            try:
                if group_label in df[group_field].values:
                    filtered_df = df.loc[df[group_field] == group_label].copy()

                    # Filter agents by index
                    filtered_df = filtered_df[
                        filtered_df.index.isin(indexes)
                        ].copy()

                    if filtered_df.shape[0] != 0:
                        # Set velocities only for the filtered_df
                        # Update df using filtered_df
                        df.update(
                            cls.set_velocities(
                                df=filtered_df,
                                distribution=distribution,
                                angle_distribution=angle_distribution
                                )
                            )
                        if preserve_dtypes_dict:
                            df = df.astype(preserve_dtypes_dict)
                else:
                    # group_label not in df[group_field].values
                    # Do nothing and return unaltered df
                    pass
            except Exception as error:
                exception_burner([
                    error,
                    check_field_existance(df, [group_field, "vx", "vy"])
                    ])
            else:
                return df
        elif indexes is not None:
            try:
                # Filter agents by index
                filtered_df = df.iloc[indexes, :].copy()

                # Set velocities only for the filtered_df
                # Update df using filtered_df
                df.update(
                    cls.set_velocities(
                        df=filtered_df,
                        distribution=distribution,
                        angle_distribution=angle_distribution
                        )
                    )
                if preserve_dtypes_dict:
                    df = df.astype(preserve_dtypes_dict)
            except Exception as error:
                exception_burner([
                    error,
                    check_field_existance(df, ["vx", "vy"])
                    ])
            else:
                return df
        elif group_field is not None:
            try:
                if group_label in df[group_field].values:
                    filtered_df = df.loc[df[group_field] == group_label].copy()

                    # Set velocities only for the filtered_df
                    # Update df using filtered_df
                    df.update(
                        cls.set_velocities(
                            df=filtered_df,
                            distribution=distribution,
                            angle_distribution=angle_distribution
                            )
                        )
                    if preserve_dtypes_dict:
                        df = df.astype(preserve_dtypes_dict)
                else:
                    # group_label not in df[group_field].values
                    # Do nothing and return unaltered df
                    pass
            except Exception as error:
                exception_burner([
                    error,
                    check_field_existance(df, [group_field, "vx", "vy"])
                    ])
            else:
                return df
        else:
            try:
                # Set velocities for all the agents in df
                df = cls.set_velocities(
                    df=df,
                    distribution=distribution,
                    angle_distribution=angle_distribution
                    )
            except Exception as error:
                exception_burner([
                    error,
                    check_field_existance(df, ["vx", "vy"])
                    ])
            else:
                return df

    @classmethod
    def update_velocities(
        cls,
        df: DataFrame,
        distribution: Distribution,
        angle_variance: float,
        indexes: Union[list, ndarray, None] = None,
        group_field: Optional[str] = None,
        group_label: Optional[str] = None,
        preserve_dtypes_dict: Optional[dict] = None
    ) -> DataFrame:
        """
            Update the velocity of a given set of agents from a given
            mobility profile (i.e. a velocity distribution) and
            deviating the resulting angles using a normal distribution
            with a standard deviation equal to `angle_variance`.

            Parameters
            ----------
            df : DataFrame
                Dataframe to apply transformation.
                Must have `vx` and `vy` columns.

            distribution : Distribution
                Mobility profile. This is the velocity distribution
                to use for updating the population velocities each
                time step.

            angle_variance : float
                Standard deviation of the normal distribution
                used for changing the direction of the velocity
                from its initial value

            indexes : TODO

            group_field : str, optional
                The field over which to filter the set of agents.
                If not provided, then the set of agents used is
                going to be the whole set of agents.

            group_label : str, optional
                The value of the `group_filed` used to filter
                the set of agents. If `group_field` is not provided,
                then this parameter is ignored.

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            The velocity direction change is calculated as:

            .. math::
                \theta_{new} = \theta_{former} + \Delta \theta

            Where :math: `\Delta \theta` is a random variable which
            follows a normal distribution with mean :math: `\mu = 0.0`
            and standard deviation equals to `angle_variance`.

            TODO: insert explanatory image

            See Also
            --------
            set_velocities : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        if indexes is not None and group_field is not None:
            try:
                if group_label in df[group_field].values:
                    filtered_df = df.loc[df[group_field] == group_label].copy()

                    # Filter agents by index
                    filtered_df = filtered_df[
                        filtered_df.index.isin(indexes)
                        ].copy()

                    if filtered_df.shape[0] != 0:
                        # Set velocities only for the filtered_df
                        # Update df using filtered_df
                        df.update(
                            cls.set_velocities(
                                df=filtered_df,
                                distribution=distribution,
                                angle_variance=angle_variance
                                )
                            )
                        if preserve_dtypes_dict:
                            df = df.astype(preserve_dtypes_dict)
                else:
                    # group_label not in df[group_field].values
                    # Do nothing and return unaltered df
                    pass
            except Exception as error:
                exception_burner([
                    error,
                    check_field_existance(df, [group_field, "vx", "vy"])
                    ])
            else:
                return df
        elif indexes is not None:
            try:
                # Filter agents by index
                filtered_df = df.iloc[indexes, :].copy()

                # Set velocities only for the filtered_df
                # Update df using filtered_df
                df.update(
                    cls.set_velocities(
                        df=filtered_df,
                        distribution=distribution,
                        angle_variance=angle_variance
                        )
                    )
                if preserve_dtypes_dict:
                    df = df.astype(preserve_dtypes_dict)
            except Exception as error:
                exception_burner([
                    error,
                    check_field_existance(df, ["vx", "vy"])
                    ])
            else:
                return df
        elif group_field is not None:
            # group_field is not None
            try:
                if group_label in df[group_field].values:
                    filtered_df = df.loc[df[group_field] == group_label].copy()

                    # Change velocities only for the filtered_df
                    # Update df using filtered_df
                    df.update(
                        cls.set_velocities(
                            df=filtered_df,
                            distribution=distribution,
                            angle_variance=angle_variance
                            )
                        )
                    if preserve_dtypes_dict:
                        df = df.astype(preserve_dtypes_dict)
                else:
                    # group_label not in df[group_field].values
                    # Do nothing and return unaltered df
                    pass
            except Exception as error:
                exception_burner([
                    error,
                    check_field_existance(df, [group_field, "vx", "vy"])
                    ])
            else:
                return df
        else:
            try:
                # Change velocities for all the agents in df
                df = cls.set_velocities(
                    df=df,
                    distribution=distribution,
                    angle_variance=angle_variance
                    )
            except Exception as error:
                exception_burner([
                    error,
                    check_field_existance(df, ["vx", "vy"])
                    ])
            else:
                return df

    @classmethod
    def deviation_angle(cls, grouped_df: DataFrame) -> float:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            standardize_angle : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        sorted_serie = \
            grouped_df["relative_angle"].sort_values().copy()

        consecutive_angle = sorted_serie.diff().shift(periods=-1)
        consecutive_angle.iloc[-1] = \
            (2*pi - sorted_serie.iloc[-1]) + sorted_serie.iloc[0]

        sorted_df = DataFrame()
        sorted_df["relative_angle"] = sorted_serie
        sorted_df["consecutive_angle"] = consecutive_angle

        max_angle = sorted_df["consecutive_angle"].max()

        greatest_angle_to_avoid = sorted_df.loc[
            sorted_df["consecutive_angle"] == max_angle
            ]
        # Random index when there are more than one max value.
        index = \
            greatest_angle_to_avoid.relative_angle.sample().index[0]
        # Standardize angles on the interval [0, 2*pi]
        return cls.standardize_angle(
            greatest_angle_to_avoid["relative_angle"].loc[index] +
            greatest_angle_to_avoid["consecutive_angle"].loc[index]/2
            )

    @classmethod
    def replace_velocities(
        cls,
        row: Series,
        new_angles: Series
    ) -> Series:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            Examples
            --------
            TODO: include some examples
        """
        if row.agent in new_angles.index:
            velocity_norm = sqrt(row.vx**2 + row.vy**2)
            row.vx = velocity_norm * cos(new_angles[row.agent])
            row.vy = velocity_norm * sin(new_angles[row.agent])
        return row

    @classmethod
    def avoid_agents(cls, df: DataFrame, df_to_avoid: DataFrame) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            deviation_angle : TODO complete explanation

            replace_velocities : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            df_copy = df.copy()

            scared_agents = df_copy.loc[df_copy.agent.isin(
                df_to_avoid["agent"].unique()
                )][["agent", "x", "y", "vx", "vy"]]

            scary_agents = scared_agents.merge(
                        df_to_avoid, how="inner", on="agent"
                        ).merge(
                    df_copy.rename(
                        columns={
                            "agent": "agent_to_avoid",
                            "x": "x_to_avoid",
                            "y": "y_to_avoid"
                            })[["agent_to_avoid", "x_to_avoid", "y_to_avoid"]],
                    how="inner",
                    on="agent_to_avoid"
                    )
            scary_agents["x_relative"] = scary_agents.apply(
                    lambda row: row.x_to_avoid - row.x, axis=1
                    )

            scary_agents["y_relative"] = scary_agents.apply(
                    lambda row: row.y_to_avoid - row.y, axis=1
                    )

            scary_agents["relative_angle"] = cls.vector_angles(
                scary_agents,
                ["x_relative", "y_relative"]
                )

            scary_agents["relative_angle"] = \
                scary_agents["relative_angle"].apply(cls.standardize_angle)

            new_angles = scary_agents[["agent", "relative_angle"]] \
                .groupby("agent").apply(cls.deviation_angle)

            df = df.apply(
                lambda row: cls.replace_velocities(row, new_angles),
                axis=1
                )
        except Exception as error:
            exception_burner([
                error,
                check_field_existance(df, ["agent", "x", "y", "vx", "vy"])
                ])
        else:
            return df



# Copyright (C) 2021, Camilo Hincapié Gutiérrez
# This file is part of CDSLIB.
#
# CDSLIB is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CDSLIB is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
#
# This package is authored by:
# Camilo Hincapié (https://www.linkedin.com/in/camilo-hincapie-gutierrez/) (main author)
# Ian Mejía (https://github.com/IanMejia)
# Emil Rueda (https://www.linkedin.com/in/emil-rueda-424012207/)
# Nicole Rivera (https://github.com/nicolerivera1)
# Carolina Rojas Duque (https://github.com/carolinarojasd)

from typing import Any

from abmodel.utils import Distribution


def init_distribution(dist_dict: dict[str, Any]) -> Distribution:
    """
        This helper function should be used for initializing
        a Distribution from `dist_dict` which should
        contain the required information.

        Parameters
        ----------
        dist_dict : dict
            Dictionary with the required information
            in order to initialize the distribution.

        Returns
        -------
        distribution : Distribution
            An initialized distribution from dist_dict.

        See Also
        --------
        abmodel.utils.distributions.Distribution : Distribution class
    """
    if dist_dict["dist_type"] == "constant":
        return Distribution(
            dist_type=dist_dict["dist_type"],
            constant=dist_dict["constant"]
            )
    elif dist_dict["dist_type"] == "empirical":
        return Distribution(
            dist_type=dist_dict["dist_type"],
            data=dist_dict["data"],
            filename=dist_dict["filename"],
            **dist_dict["kwargs"]
            )
    elif dist_dict["dist_type"] == "weights":
        return Distribution(
            dist_type=dist_dict["dist_type"],
            data=dist_dict["data"],
            filename=dist_dict["filename"],
            )
    elif dist_dict["dist_type"] == "numpy":
        return Distribution(
            dist_type=dist_dict["dist_type"],
            dist_name=dist_dict["dist_name"],
            **dist_dict["kwargs"]
            )
    else:
        return Distribution(
            dist_type=dist_dict["dist_type"]
            )



# Copyright (C) 2021, Camilo Hincapié Gutiérrez
# This file is part of CDSLIB.
#
# CDSLIB is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CDSLIB is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
#
# This package is authored by:
# Camilo Hincapié (https://www.linkedin.com/in/camilo-hincapie-gutierrez/) (main author)
# Ian Mejía (https://github.com/IanMejia)
# Emil Rueda (https://www.linkedin.com/in/emil-rueda-424012207/)
# Nicole Rivera (https://github.com/nicolerivera1)
# Carolina Rojas Duque (https://github.com/carolinarojasd)

from numpy import array, setdiff1d, concatenate, transpose
# where, full, ndarray, isin, concatenate
# from numpy.random import choice, random_sample
from pandas.core.frame import DataFrame

from abmodel.utils.execution_modes import ExecutionModes
from abmodel.utils.utilities import check_field_existance, exception_burner
from abmodel.models.disease import DiseaseStates


def trace_neighbors_vectorized(
    df: DataFrame,
    tracing_radius: float,
    kdtree_by_disease_state: dict,
    agents_labels_by_disease_state: dict,
    dead_disease_group: str,
    disease_groups: DiseaseStates
) -> DataFrame:
    """
        TODO: Add brief explanation

        Parameters
        ----------
        TODO

        Returns
        -------
        TODO

        Notes
        -----
        TODO: include mathematical description and explanatory image

        Examples
        --------
        TODO: include some examples
    """
    # Retrieve agents locations
    agents_locations = df[["x", "y"]].to_numpy()
    agents_labels = df[["agent"]].to_numpy()
    n_agents = df.shape[0]

    susceptible_neighbors = \
        [array([], dtype=int) for index in range(n_agents)]
    infected_spreader_neighbors = \
        [array([], dtype=int) for index in range(n_agents)]
    infected_non_spreader_neighbors = \
        [array([], dtype=int) for index in range(n_agents)]
    immune_neighbors = \
        [array([], dtype=int) for index in range(n_agents)]
    total_neighbors = \
        [array([], dtype=int) for index in range(n_agents)]

    # Cycle through each state of the neighbors
    for disease_state in disease_groups.items.keys():

        can_get_infected = \
            disease_groups.items[disease_state].can_get_infected

        is_infected = \
            disease_groups.items[disease_state].is_infected

        can_spread = \
            disease_groups.items[disease_state].can_spread

        if disease_state != dead_disease_group:

            if kdtree_by_disease_state[disease_state]:

                # Detect if the agents of "disease_state" that are
                # inside a distance equal to the tracing_radius
                # points_inside_radius_array is a ndarray with a list
                # of indeces which correspond neighbors of agent
                points_inside_radius_array = \
                    kdtree_by_disease_state[disease_state] \
                    .query_ball_point(
                        agents_locations,
                        tracing_radius
                        )

                # Now we have to get the corresponding agents labels
                # excluding the agent's own index
                agents_labels_inside_radius_list = [
                    setdiff1d(
                        agents_labels_by_disease_state[disease_state][
                            points_inside_radius_array[i]],
                        agents_labels[i]
                        ).astype(int)
                    for i in range(n_agents)
                    ]

                if can_get_infected:
                    # i.e. susceptibles
                    susceptible_neighbors = [
                        concatenate(
                            (susceptible_neighbors[i],
                                agents_labels_inside_radius_list[i]),
                            axis=None,
                            dtype=int
                            )
                        for i in range(n_agents)
                        ]

                if not can_get_infected and not is_infected:
                    # i.e. inmunes
                    immune_neighbors = [
                        concatenate(
                            (immune_neighbors[i],
                                agents_labels_inside_radius_list[i]),
                            axis=None,
                            dtype=int
                            )
                        for i in range(n_agents)
                        ]

                if is_infected:
                    # infected

                    if can_spread:
                        infected_spreader_neighbors = [
                            concatenate(
                                (infected_spreader_neighbors[i],
                                    agents_labels_inside_radius_list[i]),
                                axis=None,
                                dtype=int
                                )
                            for i in range(n_agents)
                            ]

                    else:
                        infected_non_spreader_neighbors = [
                            concatenate(
                                (infected_non_spreader_neighbors[i],
                                    agents_labels_inside_radius_list[i]),
                                axis=None,
                                dtype=int
                                )
                            for i in range(n_agents)
                            ]

        else:
            # disease_state == dead_disease_group
            pass

    # Extend total_neighbors
    total_neighbors = [
        concatenate(
            (total_neighbors[i],
                susceptible_neighbors[i],
                immune_neighbors[i],
                infected_spreader_neighbors[i],
                infected_non_spreader_neighbors[i]),
            axis=None,
            dtype=int
            )
        for i in range(n_agents)
        ]

    data = array([susceptible_neighbors, infected_spreader_neighbors,
                  infected_non_spreader_neighbors, immune_neighbors,
                  total_neighbors], dtype="object")
    data_transposed = transpose(data)

    return DataFrame(data_transposed)


class AgentNeighbors:
    """
        TODO: Add brief explanation

        Methods
        -------
        TODO
    """
    @classmethod
    def trace_neighbors_to_susceptibles(
        cls,
        df: DataFrame,
        tracing_radius: float,
        kdtree_by_disease_state: dict,
        agents_labels_by_disease_state: dict,
        dead_disease_group: str,
        disease_groups: DiseaseStates,
        execmode: ExecutionModes = ExecutionModes.vectorized.value
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            See Also
            --------
            trace_neighbors_vectorized : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            if execmode == ExecutionModes.vectorized.value:
                df[["susceptible_neighbors",
                    "infected_spreader_neighbors",
                    "infected_non_spreader_neighbors",
                    "immune_neighbors",
                    "total_neighbors"]] = trace_neighbors_vectorized(
                    df,
                    tracing_radius,
                    kdtree_by_disease_state,
                    agents_labels_by_disease_state,
                    dead_disease_group,
                    disease_groups
                    )
            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )
        except Exception as error:
            validation_list = ["susceptible_neighbors",
                               "infected_spreader_neighbors",
                               "infected_non_spreader_neighbors",
                               "immune_neighbors", "total_neighbors"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            return df



# Copyright (C) 2021, Camilo Hincapié Gutiérrez
# This file is part of CDSLIB.
#
# CDSLIB is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CDSLIB is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
#
# This package is authored by:
# Camilo Hincapié (https://www.linkedin.com/in/camilo-hincapie-gutierrez/) (main author)
# Ian Mejía (https://github.com/IanMejia)
# Emil Rueda (https://www.linkedin.com/in/emil-rueda-424012207/)
# Nicole Rivera (https://github.com/nicolerivera1)
# Carolina Rojas Duque (https://github.com/carolinarojasd)

from enum import Enum
from dataclasses import dataclass
from copy import deepcopy

from abmodel.models import SimpleDistGroups
from abmodel.models import ComplexDistGroups
from abmodel.utils import std_str_join_cols


# ============================================================================
# Groups with distributions in their attributes
# ============================================================================

class DistTitles(Enum):
    """
        This class enumerates some fundamental
        distribution titles that are used
        across the library.
    """
    susceptibility = "susceptibility_dist"
    mobility = "mobility_profile"
    adherence = "adherence_prob"
    diagnosis = "diagnosis_prob"
    isolation_days = "isolation_days"
    hospitalization = "hospitalization_prob"
    icu_prob = "ICU_prob"
    immunization_time = "immunization_time_distribution"
    time = "time_dist"
    alertness = "alertness_prob"
    immunization_level = "immunization_level_dist"
    mr_adherence = "mr_adherence_prob"


@dataclass
class SusceptibilityGroups(SimpleDistGroups):
    """
        Dataclass used for wrapping Susceptibility groups

        Attributes
        ----------
        dist_title : str
            Distribution title. It must be equal to
            `DistTitles.susceptibility`

        group_info : list of dict
            The list of different single group
            information required to instantiate
            a DistributionGroup.

        items : dict
            The dictionary created from the list `group_info`.
            Each key of this dictionary corresponds to a single
            group name.

        See Also
        --------
        DistTitles : fundamental distribution titles

        abmodel.models.base.SimpleDistGroups : Simple Distribution groups class

        Examples
        --------
        TODO: include some examples
    """
    def __post_init__(self):
        """
            Validates `dist_title` to be equal to
            `DistTitles.susceptibility` and then
            performs `items` dictionary
            assignment from `group_info` list.
        """
        self.single_dist_title_validation(
            expected_dist_title=DistTitles.susceptibility.value
            )
        super().__post_init__()


@dataclass
class ImmunizationGroups(SimpleDistGroups):
    """
        Dataclass used for wrapping Immunization groups

        Attributes
        ----------
        TODO

        group_info : list of dict
            The list of different single group
            information required to instantiate
            a DistributionGroup.

        items : dict
            The dictionary created from the list `group_info`.
            Each key of this dictionary corresponds to a single
            group name.

        See Also
        --------
        DistTitles : fundamental distribution titles

        abmodel.models.base.SimpleDistGroups : Simple Distribution groups class

        Examples
        --------
        TODO: include some examples
    """
    def __post_init__(self):
        """
            TODO
        """
        # TODO: Include dist_title_validation
        super().__post_init__()


@dataclass
class MobilityGroups(ComplexDistGroups):
    """
        Dataclass used for wrapping Mobility groups

        Attributes
        ----------
        dist_title : str
            Distribution title. It must be equal to
            `DistTitles.mobility`

        group_info : list of dict
            The list of different single group
            information required to instantiate
            a DistributionGroup.

        items : dict
            The dictionary created from the list `group_info`.
            Each key of this dictionary corresponds to a single
            group name.

        labels : list
            TODO complete explanation

        See Also
        --------
        DistTitles : fundamental distribution titles

        abmodel.models.base.ComplexDistGroups : TODO complete explanation

        Examples
        --------
        TODO: include some examples
    """
    def __post_init__(self):
        """
            This method performs `items` dictionary
            assignment from `group_info` list.
        """
        # TODO: Include dist_title_validation

        self.labels = [
            "angle_variance"
        ]
        super().__post_init__()


@dataclass
class IsolationAdherenceGroups(SimpleDistGroups):
    """
        Dataclass used for wrapping Adherence groups

        Attributes
        ----------
        dist_title : str
            Distribution title. It must be equal to
            `DistTitles.adherence`

        group_info : list of dict
            The list of different single group
            information required to instantiate
            a DistributionGroup.

        items : dict
            The dictionary created from the list `group_info`.
            Each key of this dictionary corresponds to a single
            group name.

        See Also
        --------
        DistTitles : fundamental distribution titles

        abmodel.models.base.SimpleDistGroups : Simple Distribution groups class

        Examples
        --------
        TODO: include some examples
    """
    def __post_init__(self):
        """
            Validates `dist_title` to be equal to
            `DistTitles.adherence` and then
            performs `items` dictionary
            assignment from `group_info` list.
        """
        self.single_dist_title_validation(
            expected_dist_title=DistTitles.adherence.value
            )
        super().__post_init__()


@dataclass
class MRAdherenceGroups(SimpleDistGroups):
    """
        Dataclass used for wrapping Adherence groups

        Attributes
        ----------
        dist_title : str
            Distribution title. It must be equal to
            `DistTitles.adherence`

        group_info : list of dict
            The list of different single group
            information required to instantiate
            a DistributionGroup.

        items : dict
            The dictionary created from the list `group_info`.
            Each key of this dictionary corresponds to a single
            group name.

        See Also
        --------
        DistTitles : fundamental distribution titles

        abmodel.models.base.SimpleDistGroups : Simple Distribution groups class

        Examples
        --------
        TODO: include some examples
    """
    def __post_init__(self):
        """
            Validates `dist_title` to be equal to
            `DistTitles.adherence` and then
            performs `items` dictionary
            assignment from `group_info` list.
        """
        self.single_dist_title_validation(
            expected_dist_title=DistTitles.mr_adherence.value
            )
        super().__post_init__()


@dataclass
class DiseaseStates(ComplexDistGroups):
    """
        TODO: Add brief explanation

        Attributes
        ----------
        TODO

        See Also
        --------
        TODO

        Examples
        --------
        TODO: include some examples
    """
    def __post_init__(self):
        """
            This method performs `items` dictionary
            assignment from `group_info` list.
        """
        # TODO: Include dist_title_validation

        self.labels = [
            "can_get_infected",
            "is_infected",
            "can_spread",
            "spread_radius",
            "spread_radius_unit",
            "spread_probability",
            "is_dead"
        ]
        super().__post_init__()

        # Validate at least one disease_state
        is_dead_list = [self.items[item].is_dead for item in self.items.keys()]

        if not all(isinstance(elem, bool) for elem in is_dead_list):
            raise ValueError("All values for is_dead must be boolean type")
        if True not in is_dead_list:
            raise ValueError("At least one value for is_dead must be True")


@dataclass
class Transitions(ComplexDistGroups):
    """
        TODO: Add brief explanation

        Attributes
        ----------
        TODO

        See Also
        --------
        TODO

        Examples
        --------
        TODO: include some examples
    """
    def __post_init__(self):
        """
            This method performs `items` dictionary
            assignment from `group_info` list.
        """
        # TODO: Include dist_title_validation

        self.labels = [
            "probability",
            "immunization_gain"
        ]

        for single_group in self.group_info:
            single_group["name"] = single_group.pop("transition_name")

        super().__post_init__()

        self.__delattr__("dist_title")
        self.__delattr__("group_info")


@dataclass
class NaturalHistory(ComplexDistGroups):
    """
        TODO: Add brief explanation

        Attributes
        ----------
        TODO

        See Also
        --------
        TODO

        Examples
        --------
        TODO: include some examples
    """
    def __post_init__(self):
        """
            This method performs `items` dictionary
            assignment from `group_info` list.
        """
        # TODO: Include dist_title_validation

        self.labels = [
            "avoidance_radius",
            "avoidance_radius_unit",
            "transition_by_contagion"
        ]

        self.transitions = {}
        for single_group in self.group_info:
            vulnerability_group = single_group.pop("vulnerability_group")
            disease_group = single_group.pop("disease_group")
            single_group["name"] = std_str_join_cols(
                str(vulnerability_group), str(disease_group)
                )

            self.transitions[single_group["name"]] = Transitions(
                dist_title=DistTitles.immunization_time,
                group_info=single_group.pop("transitions")
                )

        super().__post_init__()

        for key in self.items.keys():
            self.items[key].transitions = deepcopy(self.transitions[key].items)

        self.__delattr__("transitions")



# Copyright (C) 2021, Camilo Hincapié Gutiérrez
# This file is part of CDSLIB.
#
# CDSLIB is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CDSLIB is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
#
# This package is authored by:
# Camilo Hincapié (https://www.linkedin.com/in/camilo-hincapie-gutierrez/) (main author)
# Ian Mejía (https://github.com/IanMejia)
# Emil Rueda (https://www.linkedin.com/in/emil-rueda-424012207/)
# Nicole Rivera (https://github.com/nicolerivera1)
# Carolina Rojas Duque (https://github.com/carolinarojasd)

from typing import Union

from pydantic import validate_arguments
from pandas.core.frame import DataFrame
from pandas.core.series import Series


def check_field_existance(df: DataFrame, cols: list) -> str:
    """
        Validate wheter each column exists, if the validation fails,
        raise an error with specific information about the missing columns

        Parameters
        ----------
        df : DataFrame
            Data frame containing the positional information about the agents

        cols : list
            A list containing the columns to validate

        Returns
        -------
        True:
            If the column validation is passed.

        Raises
        ------
        ValueError
            If at least one of the columns to validate is missing.

        Examples
        --------
        TODO: include some examples
    """
    check_cols = []
    if set(cols).issubset(df.columns):
        return ""
    else:
        for col in cols:
            if not {col}.issubset(df.columns):
                check_cols.append(col)

        error_string = "df must contain: " + ", ".join(cols) + ".\n"
        check_string = ", ".join(check_cols) + " must be checked"

        raise ValueError(error_string + check_string)


def exception_burner(errors: list[str]) -> Exception:
    """
        Concatenates all error messages and raises one Exception containing all

        Parameters
        ----------
        errors : list
            List of error messages

        Raises
        ------
        Exception
            Concatenating all errors

        Examples
        --------
        TODO: include some examples
    """
    raise Exception("\n".join(errors))


def check_field_errors(df: DataFrame, debug: bool = False):
    """
        Validate if fields inside a column are correctly set, if the validation
        fails, raise an error about incorrect initialization, if debug is
        enabled, the error message contains specific information about the
        DataFrame.

        Parameters
        ----------
        df : DataFrame
            Dataframe containing the positional information about the agents

        debug : Boolean, optional
            Flag used to tell if extensive explorations over the DataFrame are
            needed

        Raises
        ------
        ValueError
            If the dataframe contains some any column with null values

        Examples
        --------
        TODO: include some examples
    """
    if df.isna().values.any():
        if debug:
            erratic_df = df[df.isna().any(axis=1)]
            error_string = "The following columns contain Null values:\n"
            error_string += ", ".join(list(erratic_df.columns))
            raise ValueError(error_string)
        else:
            error_string = (
                    "Some columns might be initialized incorrectly.\n"
                    "To pinpoint especific errors, "
                    "add `debug=True` as a parameter")
            raise ValueError(error_string)


@validate_arguments(config={"arbitrary_types_allowed": True})
def std_str_join_cols(
    col1: Union[str, Series],
    col2: Union[str, Series],
    separator: str = "-"
) -> Union[str, Series]:
    """
        TODO: brief explanation

        Parameters
        ----------
            TODO

        Raises
        ------
        ValueError
            TODO: explain the exception

        Examples
        --------
        TODO: include some examples
    """
    if type(col1) == str and type(col2) == str:
        return separator.join([col1, col2])
    elif type(col1) == Series and type(col2) == Series:
        return col1.astype(str) + separator + col2.astype(str)
    else:
        error_string = (
                "`col1` or `col2` might be provided incorrectly.\n"
                "Both `col1` and `col2` should have the same type"
                "corresponding to `str` or `pandas Series`")
        raise ValueError(error_string)
